from django.apps import AppConfig


class DjangoBlackboxConfig(AppConfig):
    """App configuration for django_blackbox."""

    default_auto_field = "django.db.models.BigAutoField"
    name = "django_blackbox"
    verbose_name = "Django Black Box"

    def ready(self):
        """Register signal handlers when app is ready."""
        # Import activity tracking to register signal handlers
        from . import activity_tracking  # noqa: F401




"""
Core services for creating incidents and building error responses.
"""
import random
import traceback
import uuid
from typing import Any

from django.db import IntegrityError
from django.http import JsonResponse
from django.utils import timezone

from django_blackbox.conf import get_conf
from django_blackbox.models import Incident
from django_blackbox.request_id import get_request_id, new_request_id
from django_blackbox.utils import (
    collect_request_meta,
    compute_signature,
    extract_ip_address,
    resolve_user,
    safe_log_to_file,
    should_capture_status_code,
)


def _json_500(message: str, incident_id: str | None, status: int = 500) -> JsonResponse:
    """
    Build a JSON 500 error response.
    
    Args:
        message: The error message to display.
        incident_id: The incident ID to include in the response.
        status: The HTTP status code.
        
    Returns:
        JsonResponse: A JSON error response.
    """
    config = get_conf()
    
    # Determine response status
    response_status = status
    if config.RETURN_400_INSTEAD_OF_500 and status >= 500:
        response_status = 400
    
    # Build response body
    if config.CUSTOM_ERROR_FORMAT:
        # Use custom error format if configured
        body = config.CUSTOM_ERROR_FORMAT.copy()
        # Replace placeholder with incident_id
        for key, value in body.items():
            if isinstance(value, str) and incident_id:
                body[key] = value.replace('<incident_id>', str(incident_id))
            if '<incident_id>' in str(value) and incident_id:
                body[key] = str(value).replace('<incident_id>', str(incident_id))
        # Add incident_id to the body
        if "incident_id" not in body and incident_id:
            body["incident_id"] = str(incident_id)
    else:
        # Default format
        body: dict[str, Any] = {"detail": message}
        
        if config.INCLUDE_INCIDENT_ID_IN_BODY and incident_id:
            body["incident_id"] = str(incident_id)
    
    return JsonResponse(body, status=response_status)


def _should_capture(
    request: Any,
    exception_class: str | None = None,
    http_status: int | None = None,
) -> bool:
    """
    Determine if an incident should be captured based on configuration.
    
    Args:
        request: The Django request object.
        exception_class: The exception class name (if any).
        http_status: The HTTP status code (if any).
        
    Returns:
        bool: True if the incident should be captured.
    """
    config = get_conf()
    
    # Check if enabled
    if not config.ENABLED:
        return False
    
    # Check if path should be ignored
    for pattern in config._compiled_ignore_paths:
        if pattern.search(request.path):
            return False
    
    # Check if exception should be ignored
    if exception_class:
        for ignored in config.IGNORE_EXCEPTIONS:
            if exception_class.startswith(ignored):
                return False
    
    # Check HTTP status (only capture 5xx)
    if http_status is not None and not (500 <= http_status < 600):
        return False
    
    # Check sampling rate
    if random.random() >= config.SAMPLE_RATE:
        return False
    
    return True


def safe_persist_incident(
    meta: dict,
    http_status: int,
    exception_class: str | None,
    exception_message: str | None,
    stacktrace: str | None,
    dedup_hash: str,
) -> Incident:
    """
    Safely persist an incident to the database with fallback logging.
    
    Args:
        meta: Request metadata dictionary.
        http_status: The HTTP status code.
        exception_class: The exception class name (or None).
        exception_message: The exception message (or None).
        stacktrace: The stacktrace (or None).
        dedup_hash: The deduplication hash.
        
    Returns:
        Incident: The created or updated incident.
    """
    config = get_conf()
    
    # Get request ID from meta or generate new one
    request_id_str = meta.get("request_id") or new_request_id()
    try:
        request_id = uuid.UUID(request_id_str)
    except (ValueError, TypeError):
        request_id = uuid.uuid4()
    
    # incident_id is generated inside create_or_increment when creating (to avoid races)
    incident_id = Incident.generate_incident_id()

    # If no stacktrace but we have an exception message that looks like a stacktrace,
    # make sure it's stored in the stacktrace field
    if not stacktrace and exception_message and len(exception_message) > 200 and "Traceback" in exception_message:
        stacktrace = exception_message

    # Prepare defaults for create_or_increment
    defaults = {
        "request_id": request_id,
        "incident_id": incident_id,
        "status": Incident.Status.OPEN,
        "http_status": http_status,
        "method": meta.get("method", "GET"),
        "path": meta.get("path", ""),
        "query_string": meta.get("query_string", ""),
        "user_id": meta.get("user_id"),
        "session_key": meta.get("session_key"),
        "ip_address": meta.get("ip_address"),
        "user_agent": meta.get("user_agent"),
        "headers": meta.get("headers", {}),
        "body_preview": meta.get("body_preview"),
        "content_type": meta.get("content_type"),
        "exception_class": exception_class,
        "exception_message": exception_message,
        "stacktrace": stacktrace,
        "occurred_at": timezone.now(),
        "dedup_hash": dedup_hash,
        "occurrence_count": 1,
    }
    
    try:
        max_attempts = 3
        for attempt in range(max_attempts):
            try:
                incident, created = Incident.objects.create_or_increment(
                    signature=dedup_hash,
                    defaults=defaults,
                    window_seconds=config.DEDUP_WINDOW_SECONDS,
                )
                return incident
            except IntegrityError as e:
                # Duplicate incident_id (race): retry; manager generates id inside transaction
                if "incident_id" in str(e) and attempt < max_attempts - 1:
                    continue
                raise
    except Exception as e:
        # Fallback logging
        import logging
        logger = logging.getLogger(__name__)
        logger.error(f"Failed to persist incident to database: {e}")
        
        # Log to file
        safe_log_to_file({
            "request_id": str(request_id),
            "incident_id": str(incident_id),
            "path": meta.get("path"),
            "http_status": http_status,
            "exception_class": exception_class,
            "exception_message": exception_message,
            "persist_error": str(e),
        })
        
        # Return a minimal incident-like object
        from types import SimpleNamespace
        incident = SimpleNamespace()
        incident.incident_id = incident_id
        incident.request_id = request_id
        return incident


def log_exception_and_build_response(request: Any, exc: Exception) -> JsonResponse | None:
    """
    Log an exception as an incident and build an appropriate response.
    
    Args:
        request: The Django request object.
        exc: The exception that occurred.
        
    Returns:
        JsonResponse | None: A JSON response if configured, None otherwise.
    """
    config = get_conf()
    
    # Check if we should capture this
    exception_class = f"{exc.__class__.__module__}.{exc.__class__.__name__}"
    if not _should_capture(request, exception_class=exception_class):
        return None
    
    # Collect metadata
    meta = collect_request_meta(request)
    
    # Capture exception details
    exception_message = str(exc) if exc else None
    stacktrace = None
    if config.CAPTURE_STACKTRACE:
        stacktrace = "".join(
            traceback.format_exception(type(exc), exc, exc.__traceback__)
        )
    
    # Compute signature
    signature = compute_signature(exception_class, meta["path"], exception_message or "")
    
    # Persist incident
    incident = safe_persist_incident(
        meta=meta,
        http_status=500,
        exception_class=exception_class,
        exception_message=exception_message,
        stacktrace=stacktrace,
        dedup_hash=signature,
    )
    
    # Mark that we've created an incident for this request
    request._django_blackbox_incident_created = True
    
    # If RETURN_400_INSTEAD_OF_500 is enabled, always return a response
    if config.RETURN_400_INSTEAD_OF_500:
        # Always return JSON response with custom format
        resp = _json_500(
            config.GENERIC_ERROR_MESSAGE, 
            incident.incident_id, 
            status=500
        )
        
        # Add headers
        if config.ADD_REQUEST_ID_HEADER and meta["request_id"]:
            resp["X-Request-ID"] = meta["request_id"]
        if config.ADD_INCIDENT_ID_HEADER:
            resp["X-Incident-ID"] = incident.incident_id
        
        return resp
    
    # Otherwise, check if we should return a JSON response
    if not config.EXPOSE_JSON_ERROR_BODY:
        return None
    
    # Check if JSON is appropriate
    accept = request.META.get("HTTP_ACCEPT", "")
    is_json_request = "application/json" in accept.lower()
    
    if not is_json_request:
        return None
    
    # Build JSON response
    resp = _json_500(config.GENERIC_ERROR_MESSAGE, incident.incident_id, status=500)
    
    # Add headers
    if config.ADD_REQUEST_ID_HEADER and meta["request_id"]:
        resp["X-Request-ID"] = meta["request_id"]
    if config.ADD_INCIDENT_ID_HEADER:
        resp["X-Incident-ID"] = incident.incident_id
    
    return resp


def log_5xx_response_and_decorate(request: Any, response: Any) -> Any:
    """
    Log a 5xx HTTP response and decorate with headers.
    
    Args:
        request: The Django request object.
        response: The HTTP response object.
        
    Returns:
        Any: The response with added headers or custom response if configured.
    """
    config = get_conf()
    
    # Check if enabled
    if not config.ENABLED:
        return response
    
    # Skip if incident was already created
    if hasattr(request, '_django_blackbox_incident_created'):
        return response
    
    # Check status code
    status_code = getattr(response, "status_code", 200)
    
    # Check if this status code should be captured based on configuration
    if not should_capture_status_code(status_code):
        return response
    
    # Check if we should capture this (sample rate, ignore paths, etc.)
    if not _should_capture(request, http_status=status_code):
        return response
    
    # Only capture if explicitly enabled
    if not config.CAPTURE_RESPONSE_5XX:
        # Still add headers
        if config.ADD_REQUEST_ID_HEADER:
            rid = get_request_id()
            if rid:
                response["X-Request-ID"] = rid
        return response
    
    # Collect metadata
    meta = collect_request_meta(request)
    
    # Try to get exception info from request (stored in process_exception)
    exception_class = None
    exception_message = f"HTTP {status_code}"
    stacktrace = None
    original_message = None
    
    # Check if we have stored exception info on the request
    if hasattr(request, '_django_blackbox_exception_info'):
        import traceback
        exc_type, exc_value, exc_tb = request._django_blackbox_exception_info
        if exc_type is not None:
            exception_class = f"{exc_type.__module__}.{exc_type.__name__}"
            exception_message = str(exc_value) if exc_value else f"HTTP {status_code}"
            original_message = exception_message
            if exc_tb is not None:
                stacktrace = "".join(traceback.format_exception(exc_type, exc_value, exc_tb))
    
    # If no stored exception, try to extract exception details from the response body
    except_type_from_request = None
    if hasattr(request, '_django_blackbox_exception_info'):
        try:
            import sys
            exc_info = sys.exc_info()
            if exc_info[0] is not None:
                except_type_from_request = f"{exc_info[0].__module__}.{exc_info[0].__name__}"
        except:
            pass
    
    try:
        # Try to get the response content
        if hasattr(response, 'data') and isinstance(response.data, dict):
            # DRF Response with data dict
            detail = response.data.get('detail', '')
            if detail and detail != config.GENERIC_ERROR_MESSAGE:
                exception_message = detail
                original_message = detail
            # Check if stacktrace was included in the response
            if 'stacktrace' in response.data:
                stacktrace = response.data['stacktrace']
        elif hasattr(response, 'content'):
            # Try to decode JSON from content
            import json
            try:
                content = response.content.decode('utf-8')
                data = json.loads(content)
                detail = data.get('detail', '')
                if detail and detail != config.GENERIC_ERROR_MESSAGE:
                    exception_message = detail
                    original_message = detail
                # Check for stacktrace in response
                if 'stacktrace' in data:
                    stacktrace = data['stacktrace']
                else:
                    # Try to get any error message from the data
                    for key in ['error', 'error_message', 'message']:
                        if key in data and data[key] and str(data[key]) != config.GENERIC_ERROR_MESSAGE:
                            exception_message = str(data[key])
                            original_message = str(data[key])
                            break
            except (UnicodeDecodeError, json.JSONDecodeError):
                # Not JSON, try as plain text
                if response.content:
                    content = response.content.decode('utf-8', errors='ignore')
                    if len(content) < 1000:  # Only use if reasonable size
                        exception_message = content
                        original_message = content
    except Exception:
        # If extraction fails, use default
        pass
    
    # Try to infer exception class from message if not already set
    if not exception_class:
        # First check if we have it from request
        if except_type_from_request:
            exception_class = except_type_from_request
        elif exception_message:
            # Try to extract the exception type from the message
            if "'" in exception_message and "object has no attribute" in exception_message:
                exception_class = "builtins.AttributeError"
            elif "does not exist" in exception_message:
                exception_class = "django.core.exceptions.ValidationError"
            # Add more patterns as needed
    
    # If we have a stacktrace but no original message, use the stacktrace as the exception message
    if stacktrace and not original_message:
        exception_message = stacktrace
    
    # Compute signature using extracted message
    signature = compute_signature(exception_class, meta["path"], exception_message)
    
    # Persist incident
    incident = safe_persist_incident(
        meta=meta,
        http_status=status_code,
        exception_class=exception_class,
        exception_message=exception_message,
        stacktrace=stacktrace,
        dedup_hash=signature,
    )
    
    # Mark that we've created an incident for this request
    request._django_blackbox_incident_created = True
    
    # If we need to return 400 instead of 500, replace the response
    if config.RETURN_400_INSTEAD_OF_500:
        # Create custom response
        resp = _json_500(
            config.GENERIC_ERROR_MESSAGE,
            incident.incident_id,
            status=500
        )
        # Add headers
        if config.ADD_REQUEST_ID_HEADER and meta["request_id"]:
            resp["X-Request-ID"] = meta["request_id"]
        if config.ADD_INCIDENT_ID_HEADER:
            resp["X-Incident-ID"] = incident.incident_id
        return resp
    
    # Otherwise just add headers
    if config.ADD_REQUEST_ID_HEADER and meta["request_id"]:
        response["X-Request-ID"] = meta["request_id"]
    if config.ADD_INCIDENT_ID_HEADER:
        response["X-Incident-ID"] = incident.incident_id
    
    return response




"""
Utility functions for redaction, hashing, metadata collection, and fallback logging.
"""
import datetime
import decimal
import hashlib
import json
import logging
import re
import traceback
import uuid
from datetime import timedelta
from ipaddress import ip_address
from typing import Any

from django.conf import settings
from django.contrib.auth import get_user_model
from django.http import HttpRequest
from django.utils import timezone

from django_blackbox.conf import get_conf
from django_blackbox.request_id import get_request_id

logger = logging.getLogger(__name__)
User = get_user_model()


def should_capture_status_code(status_code: int) -> bool:
    """
    Check if a status code should trigger incident capture.
    
    Args:
        status_code: The HTTP status code to check.
        
    Returns:
        bool: True if the status code should trigger incident capture.
    """
    config = get_conf()
    
    if not config.ENABLED:
        return False
    
    # Check if status code matches any configured capture rules
    for rule in config.CAPTURE_STATUS_CODES:
        if isinstance(rule, tuple):
            # Range: (start, end)
            start, end = rule
            if start <= status_code <= end:
                return True
        elif isinstance(rule, int):
            # Specific code
            if status_code == rule:
                return True
    
    return False


def redact_headers(headers: dict[str, Any], keys: list[str], mask: str) -> dict[str, Any]:
    """
    Redact sensitive header values.
    
    Args:
        headers: The headers dict to redact.
        keys: List of header keys to redact (case-insensitive).
        mask: The string to use as replacement.
        
    Returns:
        dict: A new dict with redacted values.
    """
    redacted = {}
    keys_lower = {k.lower() for k in keys}
    
    for key, value in headers.items():
        if key.lower() in keys_lower:
            redacted[key] = mask
        else:
            redacted[key] = value
    
    return redacted


def redact_body(
    payload: bytes | str | dict | Any,
    fields: list[str],
    mask: str,
    max_bytes: int,
    content_type: str | None,
) -> str:
    """
    Redact sensitive fields from request body and truncate.
    
    Args:
        payload: The body content (bytes, str, or dict).
        fields: List of field names to redact.
        mask: The string to use as replacement.
        max_bytes: Maximum bytes to store.
        content_type: The content type of the body.
        
    Returns:
        str: The redacted and truncated body as a string.
    """
    # Handle bytes
    if isinstance(payload, bytes):
        try:
            text = payload.decode("utf-8")
            return _redact_text_body(text, fields, mask, max_bytes, content_type)
        except UnicodeDecodeError:
            preview = payload[:max_bytes // 4]
            return f"[Binary content: {len(payload)} bytes]"
    
    # Handle string
    if isinstance(payload, str):
        return _redact_text_body(payload, fields, mask, max_bytes, content_type)
    
    # Handle dict (parsed JSON)
    if isinstance(payload, dict):
        return _redact_dict_body(payload, fields, mask, max_bytes)
    
    # Fallback
    text = str(payload)
    return text[:max_bytes] if len(text) <= max_bytes else text[:max_bytes] + "..."


def _redact_text_body(text: str, fields: list[str], mask: str, max_bytes: int, content_type: str | None) -> str:
    """Redact text body based on content type."""
    if content_type and "json" in content_type:
        try:
            obj = json.loads(text)
            if isinstance(obj, dict):
                return _redact_dict_body(obj, fields, mask, max_bytes)
        except (json.JSONDecodeError, TypeError):
            pass
    
    text_encoded = text.encode("utf-8")
    if len(text_encoded) <= max_bytes:
        return text
    
    truncated = text_encoded[:max_bytes].decode("utf-8", errors="ignore")
    return truncated + "..."


def _redact_dict_body(data: Any, fields: list[str], mask: str, max_bytes: int) -> str:
    """Recursively redact dict values and return as JSON string."""
    if not isinstance(data, dict):
        return str(data)
    
    redacted = _redact_dict_recursive(data, fields, mask)
    
    try:
        json_str = json.dumps(redacted, separators=(",", ":"))
        if len(json_str.encode("utf-8")) <= max_bytes:
            return json_str
        truncated = json_str.encode("utf-8")[:max_bytes].decode("utf-8", errors="ignore")
        return truncated + "..."
    except (TypeError, ValueError):
        return str(redacted)[:max_bytes]


def _redact_dict_recursive(obj: Any, fields: list[str], mask: str) -> Any:
    """Recursively traverse and redact dictionary values."""
    if isinstance(obj, dict):
        return {
            key: mask if key.lower() in [f.lower() for f in fields] else _redact_dict_recursive(value, fields, mask)
            for key, value in obj.items()
        }
    elif isinstance(obj, list):
        return [_redact_dict_recursive(item, fields, mask) for item in obj]
    else:
        return obj


def normalize_message(message: str) -> str:
    """
    Normalize exception messages by replacing IDs, UUIDs, and numbers with placeholders.
    
    Args:
        message: The exception message.
        
    Returns:
        str: The normalized message.
    """
    # Replace UUIDs with placeholder
    message = re.sub(
        r"[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}",
        "<UUID>",
        message,
        flags=re.IGNORECASE,
    )
    
    # Replace long numeric IDs (more than 4 digits)
    message = re.sub(r"\b\d{5,}\b", "<ID>", message)
    
    # Replace IP addresses
    try:
        ip_pattern = r"\b(?:\d{1,3}\.){3}\d{1,3}\b"
        message = re.sub(ip_pattern, "<IP>", message)
    except re.error:
        pass
    
    return message


def compute_signature(exception_class: str | None, path: str, message: str) -> str:
    """
    Compute a deduplication signature for an incident.
    
    Args:
        exception_class: The exception class name (or None).
        path: The request path.
        message: The exception message (or status code message).
        
    Returns:
        str: A SHA256 hex digest of the signature.
    """
    normalized_msg = normalize_message(message)
    signature_str = f"{exception_class or 'HTTP5xx'}|{path}|{normalized_msg}"
    return hashlib.sha256(signature_str.encode("utf-8")).hexdigest()


def extract_ip_address(request: HttpRequest) -> str | None:
    """
    Extract the client IP address from request.
    
    Prioritizes X-Forwarded-For, then X-Real-IP, then REMOTE_ADDR.
    
    Args:
        request: The Django request object.
        
    Returns:
        str | None: The IP address or None.
    """
    # Check X-Forwarded-For first
    xff = request.META.get("HTTP_X_FORWARDED_FOR")
    if xff:
        ip = xff.split(",")[0].strip()
        try:
            ip_address(ip)
            return ip
        except ValueError:
            pass
    
    # Check X-Real-IP
    xri = request.META.get("HTTP_X_REAL_IP")
    if xri:
        try:
            ip = ip_address(xri.strip())
            return str(ip)
        except ValueError:
            pass
    
    # Fallback to REMOTE_ADDR
    addr = request.META.get("REMOTE_ADDR")
    if addr:
        try:
            ip = ip_address(addr)
            return str(ip)
        except ValueError:
            pass
    
    return None


def extract_user_agent(request: HttpRequest) -> str | None:
    """
    Extract the User-Agent header.
    
    Args:
        request: The Django request object.
        
    Returns:
        str | None: The User-Agent string or None.
    """
    return request.META.get("HTTP_USER_AGENT")


def resolve_user(request: HttpRequest) -> str | None:
    """
    Resolve a user identifier from the request.
    
    Args:
        request: The Django request object.
        
    Returns:
        str | None: A string representation of the user, or None.
    """
    config = get_conf()
    
    # Try custom callable first
    if config.USER_RESOLUTION_CALLABLE:
        try:
            from django.utils.module_loading import import_string
            func = import_string(config.USER_RESOLUTION_CALLABLE)
            return func(request)
        except Exception as e:
            logger.warning(f"Error calling USER_RESOLUTION_CALLABLE: {e}")
    
    # Default: try to get authenticated user
    if hasattr(request, "user") and request.user.is_authenticated:
        return str(request.user.pk)
    
    return None


def collect_request_meta(request: HttpRequest) -> dict[str, Any]:
    """
    Collect rich metadata from the request.
    Optionally redacts sensitive data based on REDACT_SENSITIVE_DATA setting.
    
    Args:
        request: The Django request object.
        
    Returns:
        dict: A dictionary of request metadata (with optional redaction).
    """
    config = get_conf()
    
    # Collect headers
    headers = {}
    for key, value in request.META.items():
        if key.startswith("HTTP_"):
            header_name = key[5:].replace("_", "-").title()
            headers[header_name] = value
    
    # Apply redaction if configured
    if config.REDACT_SENSITIVE_DATA:
        headers = redact_headers(headers, config.REDACT_HEADERS, config.REDACT_MASK)
    
    # Parse request body if applicable
    body_preview = None
    content_type = request.content_type or request.META.get("CONTENT_TYPE", "")
    
    if hasattr(request, "body") and content_type in config.STORE_BODY_CONTENT_TYPES:
        try:
            body_bytes = request.body
            
            # Apply redaction to body if configured
            if config.REDACT_SENSITIVE_DATA:
                body_preview = redact_body(
                    body_bytes,
                    config.REDACT_FIELDS,
                    config.REDACT_MASK,
                    config.MAX_BODY_BYTES,
                    content_type,
                )
            else:
                # Store original body, just truncate to max bytes
                if len(body_bytes) > config.MAX_BODY_BYTES:
                    body_preview = body_bytes[:config.MAX_BODY_BYTES].decode("utf-8", errors="ignore") + "..."
                else:
                    body_preview = body_bytes.decode("utf-8", errors="ignore")
        except Exception as e:
            logger.warning(f"Failed to parse request body: {e}")
            body_preview = "[Unable to parse body]"
    
    return {
        "request_id": get_request_id(),
        "method": request.method,
        "path": request.path,
        "query_string": request.META.get("QUERY_STRING", ""),
        "user_id": resolve_user(request),
        "session_key": getattr(request, "session", {}).session_key if hasattr(request, "session") else None,
        "ip_address": extract_ip_address(request),
        "user_agent": extract_user_agent(request),
        "headers": headers,
        "body_preview": body_preview,
        "content_type": content_type,
    }


def safe_log_to_file(data: dict[str, Any]) -> None:
    """
    Log incident data to a fallback JSONL file if database write fails.
    
    Args:
        data: The incident data to log.
    """
    config = get_conf()
    if not config.FALLBACK_FILE_LOG:
        return
    
    try:
        log_entry = {
            "timestamp": timezone.now().isoformat(),
            **data,
        }
        
        with open(config.FALLBACK_FILE_PATH, "a") as f:
            f.write(json.dumps(log_entry, ensure_ascii=False) + "\n")
    except Exception as e:
        logger.error(f"Failed to write to fallback log file: {e}")


def sanitize_for_json(value: Any) -> Any:
    """
    Recursively convert Python objects into JSON-serializable structures.
    
    - Keeps primitives (str, int, float, bool, None) as-is.
    - Recursively processes dict, list, tuple, set.
    - Converts UUID, datetime, date, Decimal, and other unknown objects to str().
    
    Args:
        value: Any Python value to sanitize.
        
    Returns:
        JSON-serializable value (str, int, float, bool, None, dict, list).
    """
    # Primitives
    if value is None or isinstance(value, (str, int, float, bool)):
        return value
    
    # UUID / Decimal / datetime-like → string
    if isinstance(value, (uuid.UUID, decimal.Decimal, datetime.datetime, datetime.date, datetime.time, timedelta)):
        return str(value)
    
    # dict → dict
    if isinstance(value, dict):
        return {sanitize_for_json(k): sanitize_for_json(v) for k, v in value.items()}
    
    # list / tuple / set → list
    if isinstance(value, (list, tuple, set)):
        return [sanitize_for_json(v) for v in value]
    
    # Fallback: best-effort string
    try:
        return str(value)
    except Exception:
        return repr(value)



"""
Middleware for request ID tracking and 5xx error capture.
"""
import json
import random
import sys
import time
from typing import Any

from django.contrib.contenttypes.models import ContentType
from django.http.request import RawPostDataException
from django.utils.deprecation import MiddlewareMixin

from django_blackbox.conf import get_conf
from django_blackbox.models import Incident, RequestActivity
from django_blackbox.request_id import get_request_id, new_request_id, set_request_id
from django_blackbox.services import log_5xx_response_and_decorate, log_exception_and_build_response
from django_blackbox.utils import (
    extract_ip_address,
    extract_user_agent,
    redact_body,
    redact_headers,
    safe_log_to_file,
    sanitize_for_json,
)


class BodyCaptureMiddleware(MiddlewareMixin):
    """
    Middleware that safely caches the raw request body once, at the very beginning
    of the request lifecycle, so that later middlewares (like ActivityLoggingMiddleware)
    can use it without triggering RawPostDataException.
    
    NOTE: This middleware MUST be placed near the top of MIDDLEWARE, before DRF,
    CSRF, and any middleware that may read request.POST or request.body.
    
    Correct MIDDLEWARE order:
        MIDDLEWARE = [
            # ... security, sessions, etc. ...
            "django_blackbox.middleware.BodyCaptureMiddleware",       # <== Early, before DRF/CSRF
            "django_blackbox.middleware.RequestIDMiddleware",
            "django_blackbox.middleware.ActivityLoggingMiddleware",
            "django_blackbox.middleware.Capture5xxMiddleware",
            # ... DRF, authentication, etc. ...
        ]
    """

    def __init__(self, get_response):
        """Initialize middleware."""
        self.get_response = get_response
        super().__init__(get_response)

    def __call__(self, request):
        """
        Cache the raw request body for mutating methods.
        
        This reads request.body once at the very beginning, which:
        - Populates Django's internal _body cache
        - Allows DRF/CSRF to read it later without issues
        - Stores a copy in request._django_blackbox_raw_body for our logging
        """
        # Only cache for mutating methods; safe no-op for others
        raw_body = None
        try:
            if request.method.upper() in ("POST", "PUT", "PATCH", "DELETE"):
                # Accessing request.body here populates request._body and caches the stream.
                # Django will then reuse this cached body for request.POST / DRF parsing.
                raw_body = request.body
        except RawPostDataException:
            # Body was already read by an even earlier middleware; cannot recover
            raw_body = None
        except Exception:
            # Any other error; safe to ignore
            raw_body = None
        
        # Store cached raw body for downstream use
        request._django_blackbox_raw_body = raw_body
        
        response = self.get_response(request)
        return response


class RequestIDMiddleware(MiddlewareMixin):
    """
    Middleware to track request IDs and add X-Request-ID header to responses.
    
    Sets a unique request ID per request using context variables and optionally
    adds it to the response headers.
    """

    def process_request(self, request):
        """
        Set request ID from incoming header or generate new one.
        
        Args:
            request: The Django request object.
        """
        # Try to get X-Request-ID from incoming request
        incoming_rid = request.META.get("HTTP_X_REQUEST_ID")
        
        # Use incoming ID if present, otherwise generate new one
        if incoming_rid:
            request_id = incoming_rid
        else:
            request_id = new_request_id()
        
        # Set in context
        set_request_id(request_id)
        
        # Also attach to request for easy access
        request.django_blackbox_request_id = request_id

    def process_response(self, request, response):
        """
        Add X-Request-ID header to response.
        
        Args:
            request: The Django request object.
            response: The HTTP response object.
            
        Returns:
            The response with X-Request-ID header added.
        """
        config = get_conf()
        
        if config.ADD_REQUEST_ID_HEADER:
            rid = get_request_id()
            if rid:
                response["X-Request-ID"] = rid
        
        return response


class Capture5xxMiddleware(MiddlewareMixin):
    """
    Middleware to capture 5xx errors and return traceable error responses.
    
    Only captures server-side failures (5xx), never 4xx errors.
    """

    def process_exception(self, request, exception):
        """
        Process exceptions and create incidents.
        
        Args:
            request: The Django request object.
            exception: The exception that occurred.
            
        Returns:
            JsonResponse | None: A JSON response if configured, None to use default handler.
        """
        config = get_conf()
        
        if not config.ENABLED or not config.CAPTURE_EXCEPTIONS:
            return None
        
        # Store exception info in request for later retrieval
        request._django_blackbox_exception_info = sys.exc_info()
        
        # Try to log and build response
        response = log_exception_and_build_response(request, exception)
        
        # If we got a JSON response, return it
        # Otherwise, let Django's default 500 handler deal with it
        # Headers will still be added in process_response
        return response

    def process_response(self, request, response):
        """
        Capture 5xx responses and add incident headers.
        
        Args:
            request: The Django request object.
            response: The HTTP response object.
            
        Returns:
            The response with added headers.
        """
        # Skip if we already handled this in process_exception or DRF handler
        # Check if incident was already created (either via flag or X-Incident-ID header)
        if hasattr(response, '_django_blackbox_incident_created'):
            return response
            
        if hasattr(request, '_django_blackbox_incident_created') and request._django_blackbox_incident_created:
            return response
        
        # Check if X-Incident-ID header is already set (indicating DRF handler created it)
        # This is more reliable than checking the request flag
        try:
            # Works for both Django HttpResponse and DRF Response
            incident_id_header = response.get('X-Incident-ID') if hasattr(response, 'get') else None
            if not incident_id_header and hasattr(response, 'headers'):
                incident_id_header = response.headers.get('X-Incident-ID', '')
        except (AttributeError, TypeError, KeyError):
            incident_id_header = None
            
        if incident_id_header:
            # Mark request so subsequent checks also skip
            request._django_blackbox_incident_created = True
            return response
        
        response = log_5xx_response_and_decorate(request, response)
        return response


class ActivityLoggingMiddleware(MiddlewareMixin):
    """
    Middleware to log all HTTP request/response activity.
    
    Captures 2xx, 3xx, 4xx, and 5xx responses with rich metadata.
    Should be placed after RequestIDMiddleware but before Capture5xxMiddleware.
    """

    def __init__(self, get_response):
        """Initialize middleware."""
        self.get_response = get_response
        super().__init__(get_response)

    def __call__(self, request):
        """Process request and log activity."""
        from django_blackbox.activity_tracking import start_activity_context
        
        config = get_conf()
        
        # Check if activity logging is enabled
        if not config.ACTIVITY_LOG_ENABLED:
            return self.get_response(request)
        
        # Check if path should be ignored
        for pattern in config._compiled_activity_ignore_paths:
            if pattern.search(request.path):
                return self.get_response(request)
        
        # Check sample rate
        if random.random() >= config.ACTIVITY_LOG_SAMPLE_RATE:
            return self.get_response(request)
        
        # NEW: start per-request activity tracking context
        # This initializes the contextvars context that signal receivers will use
        start_activity_context()
        
        # Record start time
        start_time = time.monotonic()
        request._activity_start_time = start_time
        
        # Process request
        try:
            response = self.get_response(request)
        except Exception:
            # Even if exception occurs, try to log the activity
            response = None
            raise
        finally:
            # Always log activity, even if exception occurred
            try:
                self._log_activity(request, response, start_time)
            except Exception:
                # Never break the response due to logging errors
                import logging
                logger = logging.getLogger(__name__)
                logger.exception("Failed to log request activity")
        
        return response

    def _log_activity(self, request: Any, response: Any, start_time: float) -> None:
        """Log request activity to database."""
        config = get_conf()
        
        response_time_ms = self._get_response_time_ms(start_time)
        request_id = self._get_request_id(request)
        method, path, full_path, http_status = self._collect_basic_request_info(request, response)
        view_name, route_name = self._resolve_view_info(request)
        request_headers = self._collect_request_headers(request, config)
        request_body = self._build_request_body(request, method, config)
        response_headers = self._collect_response_headers(response, config)
        response_body = self._build_response_body(response, config)
        user, is_authenticated = self._resolve_user(request)
        ip_address = extract_ip_address(request)
        user_agent = extract_user_agent(request) or ""
        incident = self._resolve_incident(request, request_id)
        content_type, object_id = self._resolve_related_object(request)
        
        # Get manual activity context (if app code explicitly set it)
        activity_ctx = getattr(request, "_activity_change_context", None) or {}
        explicit_action = activity_ctx.get("action")
        custom_action = activity_ctx.get("custom_action", "")
        custom_payload = activity_ctx.get("custom_payload", {})
        instance_before = activity_ctx.get("instance_before", {})
        instance_after = activity_ctx.get("instance_after", {})
        instance_diff = activity_ctx.get("instance_diff", {})
        
        # NEW: overlay automatic tracked changes from signals
        # This gets instance_before/after/diff from pre_save/post_save signals
        from django_blackbox.activity_tracking import get_tracked_change_for
        tracked_before, tracked_after, tracked_diff = get_tracked_change_for(content_type, object_id)
        
        # If manual context didn't set them, use tracked values from signals
        if not instance_before and tracked_before:
            instance_before = tracked_before
        if not instance_after and tracked_after:
            instance_after = tracked_after
        if not instance_diff and tracked_diff:
            instance_diff = tracked_diff
        
        # Resolve action and finalize change context
        action, instance_before, instance_after, instance_diff, custom_action, custom_payload = \
            self._resolve_action_and_change_context(
                method,
                object_id,
                {
                    "action": explicit_action,
                    "custom_action": custom_action,
                    "custom_payload": custom_payload,
                    "instance_before": instance_before,
                    "instance_after": instance_after,
                    "instance_diff": instance_diff,
                },
            )
        
        self._create_request_activity(
            method, path, full_path,
            http_status, response_time_ms, view_name, route_name,
            request_id, incident,
            user, is_authenticated, ip_address, user_agent,
            content_type, object_id,
            request_headers, request_body,
            response_headers, response_body,
            action, instance_before, instance_after, instance_diff,
            custom_action, custom_payload,
            config,
        )
    
    def _get_response_time_ms(self, start_time: float) -> float:
        """Calculate response time in milliseconds."""
        return (time.monotonic() - start_time) * 1000
    
    def _get_request_id(self, request: Any) -> str:
        """Get request ID from request object or context."""
        return getattr(request, "django_blackbox_request_id", None) or get_request_id() or ""
    
    def _collect_basic_request_info(self, request: Any, response: Any) -> tuple[str, str, str, int]:
        """Collect basic request information."""
        method = request.method
        path = request.path
        query_string = request.META.get("QUERY_STRING", "")
        full_path = path
        if query_string:
            full_path = f"{path}?{query_string}"
        http_status = getattr(response, "status_code", 500) if response else 500
        return method, path, full_path, http_status
    
    def _resolve_view_info(self, request: Any) -> tuple[str, str]:
        """Resolve view name and route name from request."""
        view_name = ""
        route_name = ""
        if hasattr(request, "resolver_match") and request.resolver_match:
            view_name = getattr(request.resolver_match, "view_name", "") or ""
            route_name = getattr(request.resolver_match, "url_name", "") or ""
        return view_name, route_name
    
    def _collect_request_headers(self, request: Any, config: Any) -> dict:
        """Collect and redact request headers."""
        request_headers = {}
        for key, value in request.META.items():
            if key.startswith("HTTP_"):
                header_name = key[5:].replace("_", "-").title()
                request_headers[header_name] = value
        
        if config.REDACT_SENSITIVE_DATA:
            request_headers = redact_headers(request_headers, config.REDACT_HEADERS, config.REDACT_MASK)
        
        return request_headers
    
    def _build_request_body(self, request: Any, method: str, config: Any) -> str:
        """Build unified request body payload including query params and body data."""
        import logging
        from django.http import QueryDict
        
        logger = logging.getLogger(__name__)
        
        request_payload = {}
        
        # 1. Collect query parameters
        query_params = {}
        try:
            qd = request.GET
            if isinstance(qd, QueryDict) and qd:
                query_params = {
                    k: qd.getlist(k) if len(qd.getlist(k)) > 1 else qd.get(k)
                    for k in qd.keys()
                }
        except Exception:
            query_params = {}
        
        if query_params:
            request_payload["query"] = query_params
        
        # 2. Collect parsed body data (DRF request.data or request.POST)
        body_data = None
        
        # Try DRF request.data first
        if hasattr(request, "data"):
            try:
                data_attr = request.data
                if data_attr is not None:
                    body_data = data_attr
            except Exception:
                body_data = None
        
        # Fallback to request.POST if no DRF data
        if body_data is None:
            try:
                post_qd = request.POST
                if isinstance(post_qd, QueryDict) and post_qd:
                    body_data = {
                        k: post_qd.getlist(k) if len(post_qd.getlist(k)) > 1 else post_qd.get(k)
                        for k in post_qd.keys()
                    }
            except Exception:
                pass
        
        # 3. Use cached raw body from BodyCaptureMiddleware, NEVER call request.body here
        content_type_raw = request.META.get("CONTENT_TYPE", "") or ""
        content_type = content_type_raw.split(";")[0].strip().lower()
        store_body_for_methods = method.upper() in ("POST", "PUT", "PATCH", "DELETE")
        
        # Ensure application/json is always allowed
        store_body_types = set(config.STORE_BODY_CONTENT_TYPES)
        if "application/json" not in store_body_types:
            store_body_types.add("application/json")
        
        # Debug logging
        logger.debug(
            "Blackbox body debug: method=%s content_type=%s has_data_attr=%s",
            method, content_type, hasattr(request, "data"),
        )
        logger.debug(
            "Blackbox body debug: GET=%s POST=%s body_data=%s",
            dict(request.GET), dict(request.POST), body_data is not None,
        )
        raw_body_text = ""
        raw_body_cached = getattr(request, "_django_blackbox_raw_body", None)
        
        logger.debug(
            "Blackbox body debug: method=%s content_type=%s has_data_attr=%s raw_body_cached=%s",
            method, content_type, hasattr(request, "data"), raw_body_cached is not None,
        )
        
        # Try to use cached raw body if:
        # 1. We don't have parsed body data yet
        # 2. It's a mutating method
        # 3. We have a cached raw body
        # 4. Content-type is in allowed types OR it's JSON (always allow JSON)
        should_use_raw = (
            body_data is None and
            store_body_for_methods and
            raw_body_cached is not None and
            (content_type in store_body_types or "application/json" in content_type)
        )
        
        if should_use_raw and isinstance(raw_body_cached, bytes) and raw_body_cached:
            try:
                # Debug: log raw body preview
                logger.debug("Blackbox body raw cached preview: %r", raw_body_cached[:256])
            except Exception:
                pass
            
            # Try JSON first if content-type suggests it
            if "application/json" in content_type:
                try:
                    body_data = json.loads(raw_body_cached.decode("utf-8"))
                    logger.debug("Blackbox body: Successfully parsed cached raw body as JSON")
                except (json.JSONDecodeError, UnicodeDecodeError) as e:
                    # Not valid JSON; fall back to raw text
                    logger.debug("Blackbox body: JSON parse failed on cached body: %s", e)
                    truncated = raw_body_cached[:config.MAX_BODY_BYTES]
                    raw_body_text = truncated.decode("utf-8", errors="replace")
            else:
                # Not JSON content-type, store as raw text
                truncated = raw_body_cached[:config.MAX_BODY_BYTES]
                raw_body_text = truncated.decode("utf-8", errors="replace")
        
        # 4. Add body data or raw body to payload
        if body_data is not None:
            request_payload["body"] = body_data
        elif raw_body_text:
            request_payload["raw"] = raw_body_text
        
        # 5. Apply redaction if configured
        if config.REDACT_SENSITIVE_DATA and request_payload:
            if "body" in request_payload and isinstance(request_payload["body"], dict):
                redacted_body = request_payload["body"].copy()
                for field in config.REDACT_FIELDS:
                    if field in redacted_body:
                        redacted_body[field] = config.REDACT_MASK
                request_payload["body"] = redacted_body
        
        # 6. Convert to JSON string
        if not request_payload:
            return ""
        
        try:
            text = json.dumps(request_payload, default=str, separators=(",", ":"), ensure_ascii=False)
            data_bytes = text.encode("utf-8")
            if len(data_bytes) > config.MAX_BODY_BYTES:
                return data_bytes[:config.MAX_BODY_BYTES].decode("utf-8", errors="replace") + "..."
            return text
        except Exception as e:
            logger.debug("Blackbox body: Exception serializing payload: %s", e)
            return ""
    
    def _collect_response_headers(self, response: Any, config: Any) -> dict:
        """Collect and redact response headers."""
        response_headers = {}
        if response:
            try:
                if hasattr(response, "headers"):
                    # DRF Response (headers is a dict-like object)
                    response_headers = dict(response.headers)
                elif hasattr(response, "_headers"):
                    # Django HttpResponse (older style)
                    response_headers = {k: v for k, v in response._headers.values()}
                else:
                    # Try get method for standard HttpResponse
                    if hasattr(response, "keys"):
                        for key in response.keys():
                            response_headers[key] = response.get(key, "")
            except Exception:
                response_headers = {}
        
        if config.REDACT_SENSITIVE_DATA:
            response_headers = redact_headers(response_headers, config.REDACT_HEADERS, config.REDACT_MASK)
        
        return response_headers
    
    def _build_response_body(self, response: Any, config: Any) -> str:
        """Build response body string."""
        store_response_body = getattr(config, "STORE_RESPONSE_BODY", True)
        response_body = ""
        
        if store_response_body and response:
            try:
                # Try DRF Response first
                try:
                    from rest_framework.response import Response as DRFResponse
                except ImportError:
                    DRFResponse = None
                
                is_drf_response = DRFResponse is not None and isinstance(response, DRFResponse)
                has_data_attr = hasattr(response, "data")
                
                if is_drf_response or has_data_attr:
                    try:
                        data = response.data
                        if data is not None:
                            if isinstance(data, (dict, list)):
                                response_body_text = json.dumps(data, default=str, separators=(",", ":"), ensure_ascii=False)
                            else:
                                response_body_text = str(data)
                            
                            # Truncate if needed
                            response_body_bytes = response_body_text.encode("utf-8")
                            if len(response_body_bytes) > config.MAX_RESPONSE_BODY_BYTES:
                                response_body = response_body_bytes[:config.MAX_RESPONSE_BODY_BYTES].decode("utf-8", errors="replace") + "..."
                            else:
                                response_body = response_body_text
                    except (TypeError, ValueError, AttributeError):
                        # Fallback: try to stringify
                        try:
                            response_body = str(response.data)[:config.MAX_RESPONSE_BODY_BYTES]
                        except Exception:
                            response_body = ""
                
                # Fallback: use response.content if DRF data didn't work
                if not response_body and hasattr(response, "content"):
                    try:
                        raw_content = response.content
                        if isinstance(raw_content, bytes) and raw_content:
                            truncated = raw_content[:config.MAX_RESPONSE_BODY_BYTES]
                            response_body = truncated.decode("utf-8", errors="replace")
                    except Exception:
                        response_body = ""
            except Exception:
                response_body = ""
        
        return response_body
    
    def _resolve_user(self, request: Any) -> tuple[Any, bool]:
        """Resolve user from request."""
        user = None
        is_authenticated = False
        if hasattr(request, "user") and request.user.is_authenticated:
            user = request.user
            is_authenticated = True
        return user, is_authenticated
    
    def _resolve_incident(self, request: Any, request_id: str) -> Any:
        """Resolve linked incident if one was created."""
        incident = None
        if hasattr(request, "_django_blackbox_incident_created") and request._django_blackbox_incident_created:
            try:
                import uuid
                request_id_uuid = uuid.UUID(request_id) if request_id else None
                if request_id_uuid:
                    incident = Incident.objects.filter(request_id=request_id_uuid).order_by("-occurred_at").first()
            except (ValueError, TypeError, AttributeError):
                pass
        return incident
    
    def _resolve_related_object(self, request: Any) -> tuple[Any, str]:
        """
        Resolve related object (Generic FK) from request.
        
        Uses resolver_match + DRF's router semantics to find the view class and view instance,
        then resolves the model and object_id from URL kwargs.
        
        Returns:
            (content_type, object_id)
        """
        import logging
        
        logger = logging.getLogger(__name__)
        
        content_type = None
        object_id = ""
        
        match = getattr(request, "resolver_match", None)
        if not match:
            return content_type, object_id
        
        # Debug (keep it for now)
        logger.debug(
            "Blackbox related object: path=%s view_name=%s func=%r kwargs=%r",
            getattr(request, "path", ""),
            getattr(match, "view_name", ""),
            getattr(match, "func", None),
            getattr(match, "kwargs", {}),
        )
        
        try:
            # 1. Resolve view_class
            func = match.func
            view_class = None
            
            if hasattr(func, "cls"):
                # DRF ViewSet registered via router
                view_class = func.cls
            elif hasattr(func, "view_class"):
                # Django CBV
                view_class = func.view_class
            
            # 2. Resolve view_instance (DRF request)
            view_instance = None
            if hasattr(request, "parser_context") and request.parser_context:
                view_instance = request.parser_context.get("view")
            
            # 3. Resolve model
            model = None
            
            # 3a. DRF view instance queryset / serializer
            if view_instance is not None:
                qs = getattr(view_instance, "queryset", None)
                if qs is not None and hasattr(qs, "model"):
                    model = qs.model
                
                if model is None and hasattr(view_instance, "get_queryset"):
                    try:
                        qs = view_instance.get_queryset()
                        if qs is not None and hasattr(qs, "model"):
                            model = qs.model
                    except Exception:
                        pass
                
                # Try serializer.Meta.model
                if model is None and hasattr(view_instance, "get_serializer_class"):
                    try:
                        serializer_class = view_instance.get_serializer_class()
                        meta = getattr(serializer_class, "Meta", None)
                        model = getattr(meta, "model", None)
                    except Exception:
                        pass
            
            # 3b. View class-level queryset / model (for DRF generics & CBVs)
            if model is None and view_class is not None:
                qs = getattr(view_class, "queryset", None)
                if qs is not None and hasattr(qs, "model"):
                    model = qs.model
                
                if model is None and hasattr(view_class, "model"):
                    model = view_class.model
            
            # 4. Resolve object_id from URL kwargs
            kwargs = match.kwargs or {}
            if kwargs:
                # DRF semantics if view_instance exists
                if view_instance is not None:
                    lookup_field = getattr(view_instance, "lookup_field", "pk")
                    lookup_url_kwarg = getattr(view_instance, "lookup_url_kwarg", lookup_field)
                    
                    if lookup_url_kwarg in kwargs:
                        object_id = str(kwargs[lookup_url_kwarg])
                
                # Fallback to common patterns
                if not object_id:
                    for key in ["pk", "id", "object_id"]:
                        if key in kwargs:
                            object_id = str(kwargs[key])
                            break
            
            # 5. Build content_type if model is found
            if model is not None:
                try:
                    content_type = ContentType.objects.get_for_model(model)
                except Exception:
                    content_type = None
        except Exception:
            # Best-effort: never raise
            logger.debug("Failed to resolve related object", exc_info=True)
        
        return content_type, object_id
    
    def _resolve_action_and_change_context(
        self, method: str, object_id: str, activity_ctx: dict
    ) -> tuple[str, dict, dict, dict, str, dict]:
        """Resolve action and instance change context."""
        explicit_action = activity_ctx.get("action")
        custom_action = activity_ctx.get("custom_action", "")
        custom_payload = activity_ctx.get("custom_payload", {})
        instance_before = activity_ctx.get("instance_before", {})
        instance_after = activity_ctx.get("instance_after", {})
        instance_diff = activity_ctx.get("instance_diff", {})
        
        method_upper = method.upper() if method else ""
        
        if explicit_action is not None and explicit_action != "":
            action = explicit_action
        elif custom_action and (explicit_action is None or explicit_action == ""):
            action = "CUSTOM"
        else:
            # Default action mapping based on HTTP method
            if method_upper == "GET":
                action = "VIEW"
            elif method_upper == "POST":
                has_object_context = bool(object_id)
                has_instance_before = bool(instance_before)
                if has_object_context or has_instance_before:
                    action = "UPDATE"
                else:
                    action = "CREATE"
            elif method_upper in ("PUT", "PATCH"):
                action = "UPDATE"
            elif method_upper == "DELETE":
                action = "DELETE"
            else:
                action = method_upper or ""
        
        return action, instance_before, instance_after, instance_diff, custom_action, custom_payload
    
    def _create_request_activity(
        self,
        method: str, path: str, full_path: str,
        http_status: int, response_time_ms: float, view_name: str, route_name: str,
        request_id: str, incident: Any,
        user: Any, is_authenticated: bool, ip_address: Any, user_agent: str,
        content_type: Any, object_id: str,
        request_headers: dict, request_body: str,
        response_headers: dict, response_body: str,
        action: str, instance_before: dict, instance_after: dict, instance_diff: dict,
        custom_action: str, custom_payload: dict,
        config: Any,
    ) -> None:
        """Create RequestActivity record in database."""
        try:
            # Sanitize JSON fields to ensure all values are JSON-serializable
            # This converts UUID, datetime, Decimal, etc. to strings
            instance_before = sanitize_for_json(instance_before)
            instance_after = sanitize_for_json(instance_after)
            instance_diff = sanitize_for_json(instance_diff)
            custom_payload = sanitize_for_json(custom_payload)
            
            RequestActivity.objects.create(
                method=method,
                path=path,
                full_path=full_path,
                http_status=http_status,
                response_time_ms=response_time_ms,
                view_name=view_name,
                route_name=route_name,
                request_id=request_id,
                incident=incident,
                user=user,
                is_authenticated=is_authenticated,
                ip_address=ip_address,
                user_agent=user_agent,
                content_type=content_type,
                object_id=object_id,
                request_headers=request_headers,
                request_body=request_body,
                response_headers=response_headers,
                response_body=response_body,
                action=action,
                instance_before=instance_before,
                instance_after=instance_after,
                instance_diff=instance_diff,
                custom_action=custom_action,
                custom_payload=custom_payload,
            )
        except Exception as e:
            # Fallback logging
            import logging
            logger = logging.getLogger(__name__)
            logger.error(f"Failed to persist request activity to database: {e}")
            
            # Log to file
            safe_log_to_file({
                "request_id": request_id,
                "method": method,
                "path": path,
                "http_status": http_status,
                "persist_error": str(e),
            })

